@@ -11,10 +11,16 @@ import sys
 import textwrap
 import time
 import urllib
+import gettext
 
 import netcache
 import offthemes
-from offutils import is_local, looks_like_base64, looks_like_url, run, term_width, xdg
+import unmerdify
+from offutils import is_local, looks_like_base64, looks_like_url, run, term_width, xdg, _LOCALE_DIR, find_root, is_url_blocked, urlify
+
+gettext.bindtextdomain('offpunk', _LOCALE_DIR)
+gettext.textdomain('offpunk')
+_ = gettext.gettext
 
 try:
     from readability import Document
@@ -51,8 +57,8 @@ except ModuleNotFoundError:
 
 _DO_HTML = _HAS_SOUP  # and _HAS_READABILITY
 if _DO_HTML and not _HAS_READABILITY:
-    print("To improve your web experience (less cruft in webpages),")
-    print("please install python3-readability or readability-lxml")
+    print(_("To improve your web experience (less cruft in webpages),"))
+    print(_("please install python3-readability or readability-lxml"))
 
 try:
     import feedparser
@@ -74,7 +80,7 @@ if shutil.which("chafa"):
     # check for m < 1.10
     try:
         output = run("chafa --version")
-        chafa_major, chafa_minor, _ = output.split("\n")[0].split(" ")[-1].split(".")
+        chafa_major, chafa_minor, rest = output.split("\n")[0].split(" ")[-1].split(".")
         if int(chafa_major) >= 1 and int(chafa_minor) >= 10:
             _HAS_CHAFA = True
             _RENDER_IMAGE = True
@@ -91,7 +97,7 @@ if shutil.which("timg"):
         _HAS_TIMG = True
         _RENDER_IMAGE = True
 if not _RENDER_IMAGE:
-    print("To render images inline, you need either chafa >= 1.10 or timg > 1.3.2")
+    print(_("To render images inline, you need either chafa >= 1.10 or timg > 1.3.2"))
 
 # return ANSI text that can be show by less
 def inline_image(img_file, width):
@@ -109,7 +115,8 @@ def inline_image(img_file, width):
         if "image" not in mime:
             return ansi_img
     if _HAS_CHAFA:
-        inlines.append("chafa --bg white -t 1 -s %s -f symbols --animate=off")
+        # -O 0 remove optimisation and allows every line to be the same length
+        inlines.append("chafa -O 0 --bg white -t 1 -s %s -f symbols --animate=off")
     if _HAS_TIMG:
         inlines.append("timg --frames=1 -p q -g %sx1000")
     image_success = False
@@ -139,11 +146,42 @@ def terminal_image(img_file):
         except Exception as err:
             print(err)
 
+# This function returns a MIME based on the gopher selector
+# List available here:
+# gopher://spike.nagatha.fr/0/phlog/2025/2025-11-11-07-07-ChatGPT-tells-me-about-Gopher-selectors.txt 
+def get_gopher_mime(url):
+    parsed = urllib.parse.urlparse(url)
+    if parsed.scheme != "gopher":
+        mime = mimetypes.guess_type(path)[0]
+    elif len(parsed.path) >= 2:
+        itemtype = parsed.path[1]
+        path = parsed.path[2:]
+    else:
+        itemtype = "1"
+        path = ""
+    if itemtype == "0":
+        if path.endswith(".xml"):
+            mime = "application/xml"
+        else:
+            mime = "text/gemini"
+    elif itemtype == "1":
+        mime = "text/gopher"
+    elif itemtype == "h":
+        mime = "text/html"
+    elif itemtype in ("g", "I", "d","p"):
+        mime = mimetypes.guess_type(path)[0]
+    elif itemtype in ("9", "s", ";"):
+        mime = "binary"
+    elif itemtype in ("r","X"):
+        mime = "application/rss+xml"
+    else:
+        mime = "text/gopher"
+    return mime
 
 # First, we define the different content->text renderers, outside of the rest
 # (They could later be factorized in other files or replaced)
 class AbstractRenderer:
-    def __init__(self, content, url, center=True,**kwargs):
+    def __init__(self, content, url, center=True,redirects={},**kwargs):
         self.url = url
         #base url is used to construct relative urls (see <base> in html)
         self.base = None
@@ -161,6 +199,10 @@ class AbstractRenderer:
         self.options = kwargs
         # self.mime should be used only in renderer with multiple mime
         self.mime = None
+        # The library used to clean the HTML
+        self.cleanlib = _("No cleaning required")
+        #url redirections
+        self.redirects = redirects
 
     def display(self, mode=None, directdisplay=False):
         wtitle = self.get_formatted_title()
@@ -168,6 +210,12 @@ class AbstractRenderer:
             body = self.body
         else:
             body = wtitle + "\n" + self.get_body(mode=mode)
+            if "linkmode" in self.options:
+                # Avaliable linkmode are "none" and "end".
+                if self.options["linkmode"] == "end":
+                    links = self.get_links(mode=mode)
+                    for i in range(len(links)):
+                        body += "[%s] %s\n" % (i + 1, links[i])
         if directdisplay:
             print(body)
             return True
@@ -190,6 +238,9 @@ class AbstractRenderer:
     def get_theme(self):
         return self.theme
 
+    def set_redirects(self, redirects):
+        self.redirects = redirects
+
     # This class hold an internal representation of the HTML text
     class representation:
         def __init__(self, width, title=None, center=True, theme={},options={}):
@@ -455,10 +506,13 @@ class AbstractRenderer:
     def get_mode(self):
         return self.last_mode
 
+    def get_cleanlib(self):
+        return self.cleanlib
+
     def get_link(self, nb):
         links = self.get_links()
         if nb not in range(1, len(links)+1):
-            print("%s is not a valid link for %s" % (nb, self.url))
+            print(_("%s is not a valid link for %s") % (nb, self.url))
             return 0
         else:
             return links[nb - 1]
@@ -475,7 +529,7 @@ class AbstractRenderer:
             title += " (%s)" % self.get_url_title()
         return title
 
-    def get_formatted_title(self):
+    def get_formatted_title(self,linksnbr=True):
         title = self.get_url_title()
         nbr = len(self.get_links())
         if is_local(self.url):
@@ -485,7 +539,8 @@ class AbstractRenderer:
             str_last = "last accessed on %s" % time.ctime(
                 netcache.cache_last_modified(self.url)
             )
-            title += " (%s links)" % nbr
+            if linksnbr:
+                title += " (%s links)" % nbr
         return self._window_title(title, info=str_last)
 
     # this function is about creating a title derived from the URL
@@ -499,21 +554,7 @@ class AbstractRenderer:
             splitpath = self.url.split("/")
             filename = splitpath[-1]
             return filename
-        path = self.url
-        parsed = urllib.parse.urlparse(self.url)
-        red_title = parsed.hostname
-        if "user" in path:
-            i = 0
-            splitted = path.split("/")
-            while i < (len(splitted) - 1):
-                if splitted[i].startswith("user"):
-                    red_title = splitted[i + 1]
-                i += 1
-        if "~" in path:
-            for pp in path.split("/"):
-                if pp.startswith("~"):
-                    red_title = pp[1:]
-        return red_title
+        return find_root(self.url,return_value="name")
 
     # This function return a list of URL which should be downloaded
     # before displaying the page (images in HTML pages, typically)
@@ -551,17 +592,20 @@ class AbstractRenderer:
                 self.rendered_text[mode] += results[0] + "\n"
                 # we should absolutize all URLs here
                 for l in results[1]:
-                    ll = l.split()[0]
-                    try:
-                        abs_l = urllib.parse.urljoin(self.url, ll)
-                        self.links[mode].append(abs_l)
-                    except Exception:
-                        print(
-                            "Urljoin Error: Could not make an URL out of %s and %s"
-                            % (self.url, ll)
-                        )
-                for l in self.get_subscribe_links()[1:]:
-                    self.links[mode].append(l[0])
+                    ll = l.split()
+                    if len(ll) > 0:
+                        try:
+                            abs_l = urllib.parse.urljoin(self.url, ll[0])
+                        except Exception:
+                            print(_(
+                                "Urljoin Error: Could not make an URL out of %s and %s"
+                                % (self.url, ll)
+                                ))
+                    else:
+                        abs_l = self.url
+                    self.links[mode].append(abs_l)
+                #for l in self.get_subscribe_links()[1:]:
+                #    self.links[mode].append(l[0])
 
     def get_body(self, width=None, mode=None):
         if not mode:
@@ -738,6 +782,8 @@ class GemtextRenderer(AbstractRenderer):
                     links.append(strippedline)
                     splitted = strippedline.split(maxsplit=1)
                     url = splitted[0]
+                    # We join with current root in case it is relative
+                    abs_url = urllib.parse.urljoin(self.url, url )
                     name = None
                     if len(splitted) > 1:
                         name = splitted[1]
@@ -746,7 +792,47 @@ class GemtextRenderer(AbstractRenderer):
                     # 600 seconds after this page, we consider it as a new_link
                     current_modif = netcache.cache_last_modified(self.url)
                     link_modif = netcache.cache_last_modified(url)
+                    # Let’s see first if this is a picture
+                    image_displayed = False
                     if (
+                        _RENDER_IMAGE
+                        and not self.url.startswith("list://")
+                        # check if images are enabled in Gemini!
+                        and "gemini_images" in self.options.keys()
+                        and self.options["gemini_images"]
+                        # Check that it looks like an image
+                       # and link_modif  # There’s a valid cache for link target 
+                        and url[-4:].lower() in [".jpg",".png",".gif","jpeg"] 
+                        and netcache.is_cache_valid(abs_url)
+                    ):
+                        ansi_img = ""
+                        try:
+                            # 4 followings line are there to translate the URL into cache path
+                            img = netcache.get_cache_path(abs_url)
+                            renderer = ImageRenderer(img, abs_url)
+                            # Image width is set in the option to 40 by default
+                            # it cannot be bigger than the width of the text
+                            if "images_size" in self.options.keys() and width and \
+                                                width > self.options["images_size"] :
+                                size = self.options["images_size"]
+                            else:
+                                size = width
+                            ansi_img += renderer.get_body(width=size, mode="inline")
+                            image_displayed = True
+                        except Exception as err:
+                            # we sometimes encounter really bad formatted files or URL
+                            # we fall back to normal links in that case
+                            image_displayed = False
+                        r.add_block(ansi_img)
+                        r.open_theme("image_link")
+                        r.center_line()
+                        theme = "image_link"
+                    #theme for blocked URL
+                    elif is_url_blocked(url,self.redirects) \
+                         and r.open_theme("blocked_link"):
+                        theme = "blocked_link"
+                    #theme for recently updated URL
+                    elif (
                         current_modif
                         and link_modif
                         and current_modif - link_modif < 600
@@ -840,7 +926,7 @@ class GopherRenderer(AbstractRenderer):
                 body, width=width, mode=mode, startlinks=startlinks
             )
         except Exception as err:
-            print("Error rendering Gopher ", err)
+            print(_("Error rendering Gopher "), err)
             r = self.representation(width, theme=self.theme,options=self.options)
             r.add_block(body)
             render = r.get_final()
@@ -871,7 +957,7 @@ class GopherRenderer(AbstractRenderer):
                     # If line starts with TAB, there’s no name.
                     # We thus hide this line
                     if name:
-                        itemtype = name[0]
+                        itemtype = name[0].strip("/")
                         name = name[1:]
                         if port == "70":
                             port = ""
@@ -880,13 +966,37 @@ class GopherRenderer(AbstractRenderer):
                         if itemtype == "h" and path.startswith("URL:"):
                             url = path[4:]
                         else:
+                            # some gophermap lines include a selector without a leading "/"
+                            # gopher://some.domain/1phlog/ is valid
+                            # this is perfectly valid, and offpunk shouldn't modify the selectors
+                            # if not path.startswith("/") and itemtype:
+                            #     path = "/" + path
                             url = "gopher://%s%s/%s%s" % (host, port, itemtype, path)
-                        url = url.replace(" ", "%20")
                         linkline = url + " " + name
                         links.append(linkline)
                         number = len(links) + startlinks
-                        towrap = "[%s] " % str(number) + name
+                        protocol = ""
+                        if not url.startswith("gopher"):
+                            protocol = " " + url.split("://")[0]
+                        towrap = "[%s%s] " % (str(number), protocol) + name
+                        # If the link point to a page that has been cached less than
+                        # 600 seconds after this page, we consider it as a new_link
+                        current_modif = netcache.cache_last_modified(self.url)
+                        link_modif = netcache.cache_last_modified(url)
+                        if (
+                            current_modif
+                            and link_modif
+                            and current_modif - link_modif < 600
+                            and r.open_theme("new_link")
+                        ):
+                            theme = "new_link"
+                        elif r.open_theme("oneline_link"):
+                            theme = "oneline_link"
+                        else:
+                            theme = "link"
+                            r.open_theme("link")
                         r.add_text(towrap)
+                        r.close_theme(theme)
                 else:
                     r.add_text(line)
         return r.get_final(), links
@@ -931,8 +1041,10 @@ class FolderRenderer(GemtextRenderer):
             listfiles = os.listdir(listdir)
             if len(listfiles) > 0:
                 for l in listfiles:
-                    # removing the .gmi at the end of the name
-                    lists.append(l[:-4])
+                    #We only take gmi files
+                    if l.endswith(".gmi"):
+                        # removing the .gmi at the end of the name
+                        lists.append(l[:-4])
         if len(lists) > 0:
             body = ""
             my_lists = []
@@ -953,16 +1065,16 @@ class FolderRenderer(GemtextRenderer):
                     else:
                         my_lists.append(l)
             if len(my_lists) > 0:
-                body += "\n## Bookmarks Lists (updated during sync)\n"
+                body += _("\n## Bookmarks Lists (updated during sync)\n")
                 body += write_list(my_lists)
             if len(subscriptions) > 0:
-                body += "\n## Subscriptions (new links in those are added to tour)\n"
+                body += _("\n## Subscriptions (new links in those are added to tour)\n")
                 body += write_list(subscriptions)
             if len(frozen) > 0:
-                body += "\n## Frozen (fetched but never updated)\n"
+                body += _("\n## Frozen (fetched but never updated)\n")
                 body += write_list(frozen)
             if len(system_lists) > 0:
-                body += "\n## System Lists\n"
+                body += _("\n## System Lists\n")
                 body += write_list(system_lists)
             return [[body, None]]
 
@@ -982,10 +1094,16 @@ class FeedRenderer(GemtextRenderer):
         if not parsed:
             return False
         elif parsed.bozo:
+            #print("bozo "+str(parsed.bozo_exception))
             return False
         else:
-            # If no content, then fallback to HTML
-            return len(parsed.entries) > 0
+            # If the second element is <rss, no doubt
+            if self.body.split(">")[1].startswith("<rss"):
+                return True
+            # If not, we may suspect a HTML file if there
+            # is no entry in the ffed
+            else:
+                return len(parsed.entries) > 0
 
     def get_title(self):
         if not self.title:
@@ -1098,6 +1216,21 @@ class ImageRenderer(AbstractRenderer):
         # Now centering the image
         lines = ansi_img.splitlines()
         new_img = ""
+        # What if the picture is smaller than requested width?
+        # We try to measure it with the number of "m" or " " in the longest line.
+        # when not optimized, there are always 2 m per symbols
+        # Yes, this is a naughty ANSI hack
+        longestline = 0
+        for l in lines:
+            linelength = l.count("[0m")
+            #print("DEBUG: linelength "+str(linelength))
+            if linelength > longestline: longestline = linelength
+        #print("DEBUG: longestline: "+str(longestline))
+        newspaces = (width - longestline) //2 + 1
+        #print("DEBUG: newspaces: "+str(newspaces))
+        if newspaces > spaces and longestline < width:
+            spaces = newspaces
+        #print("DEBUG: spaces: "+str(spaces))
         for l in lines:
             new_img += spaces * " " + l + "\n"
         return new_img, []
@@ -1122,27 +1255,29 @@ class HtmlRenderer(AbstractRenderer):
     def is_valid(self):
         if not _DO_HTML:
             print(
-                "HTML document detected. Please install python-bs4 and python-readability."
+                _("HTML document detected. Please install python-bs4 and python-readability.")
             )
         return _DO_HTML and self.validity
 
     def get_subscribe_links(self):
-        subs = [[self.url, self.get_mime(), self.get_title()]]
-        soup = BeautifulSoup(self.body, "html.parser")
-        links = soup.find_all("link", rel="alternate", recursive=True)
-        for l in links:
-            ty = l.get("type")
-            if ty:
-                if "rss" in ty or "atom" in ty or "feed" in ty:
-                    # some rss links are relatives: we absolutise_url
-                    sublink = urllib.parse.urljoin(self.url, l.get("href"))
-                    subs.append([sublink, ty, l.get("title")])
+        subs = []
+        if _DO_HTML :
+            subs = [[self.url, self.get_mime(), self.get_title()]]
+            soup = BeautifulSoup(self.body, "html.parser")
+            links = soup.find_all("link", rel="alternate", recursive=True)
+            for l in links:
+                ty = l.get("type")
+                if ty:
+                    if "rss" in ty or "atom" in ty or "feed" in ty:
+                        # some rss links are relatives: we absolutise_url
+                        sublink = urllib.parse.urljoin(self.url, l.get("href"))
+                        subs.append([sublink, ty, l.get("title")])
         return subs
 
     def get_title(self):
         if self.title:
             return self.title
-        elif self.body:
+        elif _DO_HTML and self.body:
             if _HAS_READABILITY:
                 try:
                     readable = Document(self.body)
@@ -1161,7 +1296,7 @@ class HtmlRenderer(AbstractRenderer):
 
     def get_base_url(self):
         if not self.base :
-            if self.body:
+            if _DO_HTML and self.body:
                 soup = BeautifulSoup(self.body, "html.parser")
                 if soup.base :
                     base = soup.base.get("href")
@@ -1181,7 +1316,7 @@ class HtmlRenderer(AbstractRenderer):
             width = term_width()
         if not _DO_HTML:
             print(
-                "HTML document detected. Please install python-bs4 and python-readability."
+                _("HTML document detected. Please install python-bs4 and python-readability.")
             )
             return
         # This method recursively parse the HTML
@@ -1230,6 +1365,8 @@ class HtmlRenderer(AbstractRenderer):
             # never start with a "\n"
             # string = string.lstrip("\n")
             string = string.replace("\r", "").replace("\n", " ").replace("\t", " ")
+            #now we replace the rarely found Start of guarded area
+            string = string.replace("\x96","–").replace("\x91","'")
             # remove soft hyphens
             string = string.replace("\u00ad","")
             endspace = string.endswith(" ") or string.endswith("\xa0")
@@ -1323,25 +1460,48 @@ class HtmlRenderer(AbstractRenderer):
                 link = element.get("href")
                 # support for images nested in links
                 if link:
+                    # First, we transform any space that can be found in that link
+                    link = urlify(link)
                     text = ""
                     imgtext = ""
+                    #normal link, not image
+                    normal_link = True
+                    #display link to the picture?
+                    display_this_picture = False
                     # we display images first in a link
                     for child in element.children:
                         if child.name == "img":
                             recursive_render(child)
-                            imgtext = "[IMG LINK %s]"
-                    links.append(link + " " + text)
-                    link_id = str(len(links) + startlinks)
-                    r.open_theme("link")
+                            display_this_picture = True
+                            normal_link = False
+                            #print( "%s same as previous link %s ?"%(link,str(links[-1])))
+                    # we check if the link is the same as the image itself
+                    # if so, it is the last link in the links list
+                    abs_url = urllib.parse.urljoin(self.get_base_url(),link)
+                    if not normal_link and len(links) > 0:
+                        last_link = links[-1].split()
+                        if len(last_link) > 0:
+                            display_this_picture = abs_url != last_link[0]
+                    if display_this_picture:
+                        imgtext = "[IMG LINK %s]"
+                    if display_this_picture or normal_link:
+                        links.append(link + " " + text)
+                        link_id = str(len(links) + startlinks)
+                    if is_url_blocked(link,self.redirects) \
+                        and r.open_theme("blocked_link"):
+                        linktheme = "blocked_link"
+                    else:
+                        linktheme = "link"
+                        r.open_theme(linktheme)
                     for child in element.children:
                         if child.name != "img":
                             recursive_render(child, preformatted=preformatted)
-                    if imgtext != "":
+                    if display_this_picture:
                         r.center_line()
                         r.add_text(imgtext % link_id)
-                    else:
+                    elif normal_link:
                         r.add_text(" [%s]" % link_id)
-                    r.close_theme("link")
+                    r.close_theme(linktheme)
                 else:
                     # No real link found
                     for child in element.children:
@@ -1349,7 +1509,6 @@ class HtmlRenderer(AbstractRenderer):
             elif element.name == "img":
                 src = element.get("src")
                 text = ""
-                ansi_img = render_image(src, width=width, mode=mode)
                 alt = element.get("alt")
                 if alt:
                     alt = sanitize_string(alt)
@@ -1362,7 +1521,13 @@ class HtmlRenderer(AbstractRenderer):
                     abs_url, data = looks_like_base64(src, self.get_base_url())
                     # if abs_url is None, it means we don’t support
                     # the image (such as svg+xml). So we hide it.
+                    # But we first check if there’s a data-src
+                    if not abs_url:
+                        src = element.get("data-src")
+                        abs_url, data = looks_like_base64(src,self.get_base_url())
                     if abs_url:
+                        ansi_img = render_image(src, width=width, mode=mode)
+                        abs_url = urlify(abs_url)
                         links.append(abs_url + " " + text)
                         self.images[mode].append(abs_url)
                         link_id = " [%s]" % (len(links) + startlinks)
@@ -1442,18 +1607,40 @@ class HtmlRenderer(AbstractRenderer):
                         recursive_render(child, indent=indent)
 
         # the real render_html hearth
-        # note to vjousse: use self.options["ftr_site_config"] here
-        # Use self.url to mach against an url
+        # We will transform the body into a "summary" (clean-up version)
+        summary = None
+        self.cleanlib = ""
+        # if mode full, we don’t clean anything
         if mode in ["full", "full_links_only"]:
             summary = body
-        elif _HAS_READABILITY:
-            try:
-                readable = Document(body)
-                summary = readable.summary()
-            except Exception:
+            self.cleanlib += "Full as requested"
+        # let’s try unmerdify
+        elif "ftr_site_config" in self.options.keys() and self.options["ftr_site_config"]:
+            ftr = ftr_site_config=self.options["ftr_site_config"]
+            # we want to unmerdify only if there’s a rule
+            if unmerdify.is_unmerdifiable(self.url,ftr):
+                try:
+                    summary = unmerdify.unmerdify_html(body,url=self.url,\
+                            ftr_site_config=ftr,NOCONF_FAIL=False)
+                except Exception as e:
+                    self.cleanlib += "Unmerdify CRASH - %s - "%e
+                if not summary:
+                    self.cleanlib += "Unmerdify failed - returns empty html"
+                else:
+                    self.cleanlib += "Unmerdify"
+        if not summary:
+            # if no summary from unmerdify, we try readabilitty
+            if _HAS_READABILITY:
+                try:
+                    readable = Document(body)
+                    summary = readable.summary()
+                    self.cleanlib += " - Readability"
+                except Exception as e:
+                    summary = body
+                    self.cleanlib += " - Full (Readability failed) %s"%e
+            else:
                 summary = body
-        else:
-            summary = body
+                self.cleanlib += " - Full (No readability installed)"
         soup = BeautifulSoup(summary, "html.parser")
         # soup = BeautifulSoup(summary, 'html5lib')
         if soup:
@@ -1461,6 +1648,27 @@ class HtmlRenderer(AbstractRenderer):
                 recursive_render(soup.body)
             else:
                 recursive_render(soup)
+        # inserting available feeds at the end of the page (if any)
+        sublinks = self.get_subscribe_links()
+        if len(sublinks) > 1:
+            r.newparagraph()
+            r.open_theme("subtitle")
+            r.add_text("Available feeds: ")
+            r.close_theme("subtitle")
+            r.newparagraph()
+        for s in sublinks[1:]:
+            title = str(s[2])
+            mime = str(s[1])
+            # we remove the "application/" part of the mime
+            if "/" in mime: mime = mime.split("/")[1]
+            text = title + " (%s)"%mime
+            url = str(s[0])
+            links.append(url + " " + text)
+            link_id = str(len(links) + startlinks)
+            r.open_theme("link")
+            r.add_text("%s [%s]" %(text,link_id))
+            r.close_theme("link")
+            r.newline()
         return r.get_final(), links
 
 ## Now the custom renderers
@@ -1474,14 +1682,17 @@ class XkcdRenderer(HtmlRenderer):
     def has_direct_display(self):
         return _RENDER_IMAGE
 
-    #Custom renderer should return false when they can’t handle a specific content
-    #This allow fallback to normal html renderer
-    def is_valid(self):
+    def get_xkcd_number(self):
         path = urllib.parse.urlparse(self.url).path
         #We strip the leading "/" to avoid empty elements in splitted
         splitted = path.strip("/").split("/")
         # Custom renderer only works for pure comics which have alphanumeric paths 
-        return splitted[0].isalnum()
+        return splitted[0]
+
+    #Custom renderer should return false when they can’t handle a specific content
+    #This allow fallback to normal html renderer
+    def is_valid(self):
+        return self.get_xkcd_number().isalnum()
 
     def get_images(self, mode="readable"):
         img_url,img_path,alttext,title = self.xkcd_extract()
@@ -1496,37 +1707,40 @@ class XkcdRenderer(HtmlRenderer):
 
     #return [image_url,image_path, image_alt_text,image_title]
     def xkcd_extract(self):
-        soup = BeautifulSoup(self.body, "html.parser")
-        comic_div = soup.find("div",{"id":"comic"})
-        if comic_div:
-            img_element = comic_div.find("img")
-            src=img_element.get("src")
-            if src.startswith("//"):
-                scheme = urllib.parse.urlparse(self.url).scheme
-                img_url = scheme + ":" + src
-            else:
-                img_url = src
-            img_path = netcache.get_cache_path(img_url)
-            alttext=img_element.get("title")
-            title=img_element.get("alt")
-            return img_url,img_path,alttext,title
-        else:
-            return None,None,None,None
+        if _DO_HTML :
+            soup = BeautifulSoup(self.body, "html.parser")
+            comic_div = soup.find("div",{"id":"comic"})
+            if comic_div:
+                img_element = comic_div.find("img")
+                src=img_element.get("src")
+                if src.startswith("//"):
+                    scheme = urllib.parse.urlparse(self.url).scheme
+                    img_url = scheme + ":" + src
+                else:
+                    img_url = src
+                img_path = netcache.get_cache_path(img_url)
+                alttext=img_element.get("title")
+                title=img_element.get("alt")
+                return img_url,img_path,alttext,title
+        return None,None,None,None
 
     def display(self, mode=None, directdisplay=False):
-        wtitle = self.get_formatted_title()
+        info = "  (XKCD #%s)"%self.get_xkcd_number()
+        wtitle = self.get_formatted_title(linksnbr=False)
         if not directdisplay:
             body = wtitle + "\n" + self.get_body(mode=mode)
             return body
         else:
-            print(self._window_title(wtitle))
-            self.printgemtext("# "+self.get_title())
+            print(wtitle)
+            self.printgemtext("# "+self.get_title() + info)
             img_url,img_path,alttext,title = self.xkcd_extract()
             #now displaying
             if img_path and netcache.is_cache_valid(img_url):
                 terminal_image(img_path)
+            elif not _DO_HTML:
+                self.printgemtext(_("\n> Please install python-bs4 to parse HTML"))
             else:
-                self.printgemtext("\n> missing picture, please reload\n")
+                self.printgemtext(_("\n> Picture not in cache. Please reload this page.\n"))
             self.printgemtext(alttext)
             return True
 
@@ -1549,6 +1763,8 @@ _FORMAT_RENDERERS = {
     "text/empty": EmptyRenderer,
     "message/news": GemtextRenderer,
     "message/rfc822": GemtextRenderer,
+    "application/pgp-keys": PlaintextRenderer,
+    "application/pgp-signature": PlaintextRenderer,
 }
 
 _CUSTOM_RENDERERS = {
@@ -1565,29 +1781,7 @@ def get_mime(path, url=None):
         return "text/empty"
     elif url and url.startswith("gopher://"):
         # special case for gopher
-        # code copy/pasted from netcache
-        parsed = urllib.parse.urlparse(url)
-        if len(parsed.path) >= 2:
-            itemtype = parsed.path[1]
-            path = parsed.path[2:]
-        else:
-            itemtype = "1"
-            path = ""
-        if itemtype == "0":
-            if path.endswith(".xml"):
-                mime = "application/xml"
-            else:
-                mime = "text/gemini"
-        elif itemtype == "1":
-            mime = "text/gopher"
-        elif itemtype == "h":
-            mime = "text/html"
-        elif itemtype in ("g", "I"):
-            mime = mimetypes.guess_type(path)[0]
-        elif itemtype in ("9", "s", ";"):
-            mime = "binary"
-        else:
-            mime = "text/gopher"
+        mime = get_gopher_mime(url)
     elif path.startswith("mailto:"):
         mime = "mailto"
     elif os.path.isdir(path):
@@ -1629,7 +1823,7 @@ def get_mime(path, url=None):
         mime, encoding = mimetypes.guess_type(path, strict=False)
     # gmi Mimetype is not recognized yet
     if not mime and not shutil.which("file"):
-        print('Cannot guess the mime type of the file. Please install "file".')
+        print(_('Cannot guess the mime type of the file. Please install "file".'))
     if mime.startswith("text") and mime not in _FORMAT_RENDERERS:
         if mime2 and mime2 in _FORMAT_RENDERERS:
             mime = mime2
@@ -1643,7 +1837,7 @@ def get_mime(path, url=None):
     return mime
 
 
-def renderer_from_file(path, url=None, theme=None,**kwargs):
+def renderer_from_file(path, url=None, theme=None, redirectlist={}, **kwargs):
     if not path:
         return None
     mime = get_mime(path, url=url)
@@ -1656,14 +1850,14 @@ def renderer_from_file(path, url=None, theme=None,**kwargs):
                 f.close()
         else:
             content = path
-        toreturn = set_renderer(content, url, mime, theme=theme\
-                                ,**kwargs)
+        toreturn = set_renderer(content, url, mime, theme=theme, \
+                                redirectlist=redirectlist,**kwargs)
     else:
         toreturn = None
     return toreturn
 
 
-def set_renderer(content, url, mime, theme=None,**kwargs):
+def set_renderer(content, url, mime, theme=None, redirectlist={}, **kwargs):
     renderer = None
     if mime == "Local Folder":
         renderer = FolderRenderer("", url, datadir=xdg("data"))
@@ -1706,13 +1900,16 @@ def set_renderer(content, url, mime, theme=None,**kwargs):
     if not renderer:
         renderer = FakeRenderer("",url,**kwargs)
         renderer.set_mime(mime)
-    if renderer and theme:
-        renderer.set_theme(theme)
+    if renderer:
+        if theme:
+            renderer.set_theme(theme)
+        if redirectlist:
+            renderer.set_redirects(redirectlist)
     return renderer
 
 
 # This function should be removed and replaced by a set_renderer()/r.display()
-def render(input, path=None, format="auto", mime=None, url=None, mode=None):
+def render(input, path=None, format="auto", mime=None, url=None, mode=None, linkmode="none"):
     if not url:
         url = ""
     else:
@@ -1737,13 +1934,14 @@ def render(input, path=None, format="auto", mime=None, url=None, mode=None):
         else:
             r = set_renderer(input, url, mime)
     if r:
+        r.options["linkmode"] = linkmode
         r.display(directdisplay=True, mode=mode)
     else:
-        print("Could not render %s" % input)
+        print(_("Could not render %s") % input)
 
 
 def main():
-    descri = (
+    descri = _(
         "ansicat is a terminal rendering tool that will render multiple formats (HTML, \
             Gemtext, RSS, Gophermap, Image) into ANSI text and colors.\n\
             When used on a file, ansicat will try to autodetect the format. When used with \
@@ -1765,19 +1963,27 @@ def main():
             "text",
             "plaintext",
         ],
-        help="Renderer to use. Available: auto, gemtext, html, feed, gopher, image, folder, plaintext",
+        help=_("Renderer to use. Available: auto, gemtext, html, feed, gopher, image, folder, plaintext"),
     )
-    parser.add_argument("--mime", help="Mime of the content to parse")
+    parser.add_argument("--mime", help=_("Mime of the content to parse"))
     ## The argument needs to be a path to a file. If none, then stdin is used which allows
     ## to pipe text directly into ansirenderer
     parser.add_argument(
-        "--url", metavar="URL", nargs="*", help="Original URL of the content"
+        "--url", metavar="URL", nargs="*", help=_("Original URL of the content")
     )
     parser.add_argument(
         "--mode",
         metavar="MODE",
-        help="Which mode should be used to render: normal (default), full or source.\
-                                With HTML, the normal mode try to extract the article.",
+        help=_("Which mode should be used to render: normal (default), full or source.\
+                                With HTML, the normal mode try to extract the article."),
+    )
+    parser.add_argument(
+        "--linkmode",
+        choices=[
+            "none",
+            "end",
+        ],
+        help=_("Which mode should be used to render links: none (default) or end"),
     )
     parser.add_argument(
         "content",
@@ -1785,7 +1991,7 @@ def main():
         nargs="*",
         type=argparse.FileType("r"),
         default=sys.stdin,
-        help="Path to the text to render (default to stdin)",
+        help=_("Path to the text to render (default to stdin)"),
     )
     args = parser.parse_args()
     # Detect if we are running interactively or in a pipe
@@ -1805,13 +2011,14 @@ def main():
                     url=args.url,
                     mime=args.mime,
                     mode=args.mode,
+                    linkmode=args.linkmode,
                 )
         else:
-            print("Ansicat needs at least one file as an argument")
+            print(_("Ansicat needs at least one file as an argument"))
     else:
         # we are in stdin
         if not args.format and not args.mime:
-            print("Format or mime should be specified when running with stdin")
+            print(_("Format or mime should be specified when running with stdin"))
         else:
             render(
                 args.content.read(),
@@ -1820,6 +2027,7 @@ def main():
                 url=args.url,
                 mime=args.mime,
                 mode=args.mode,
+                linkmode=args.linkmode,
             )
 
 
