@@ -11,12 +11,19 @@ import ssl
 import sys
 import time
 import urllib.parse
+import http.cookiejar
 import warnings
 from ssl import CertificateError
+import gettext
 
 import ansicat
 import offutils
-from offutils import xdg
+from offutils import xdg, _LOCALE_DIR, get_url_redirected
+import offblocklist
+
+gettext.bindtextdomain('offpunk', _LOCALE_DIR)
+gettext.textdomain('offpunk')
+_ = gettext.gettext
 
 try:
     import chardet
@@ -116,7 +123,7 @@ def is_cache_valid(url, validity=0):
         # If path is too long, we always return True to avoid
         # fetching it.
         if len(cache) > 259:
-            print("We return False because path is too long")
+            print(_("We return False because path is too long"))
             return False
         if os.path.exists(cache) and not os.path.isdir(cache):
             if validity > 0:
@@ -133,11 +140,14 @@ def is_cache_valid(url, validity=0):
         # There’s not even a cache!
         return False
 
-
-def get_cache_path(url, add_index=True):
+# get_cache_path gives a local path unique to each URL
+def get_cache_path(url, add_index=True, include_protocol=True, xdgfolder="cache",subfolder=""):
     # Sometimes, cache_path became a folder! (which happens for index.html/index.gmi)
     # In that case, we need to reconstruct it
     # if add_index=False, we don’t add that "index.gmi" at the ends of the cache_path
+    # include protocol will include the part before :// in the path
+    # xdgfolder is the XDG folder to be used
+    # subfolder is the optional folder inside XDG/offpunk/
     # First, we parse the URL
     if not url:
         return None
@@ -179,21 +189,17 @@ def get_cache_path(url, add_index=True):
         # special gopher selector case
         if scheme == "gopher":
             if len(parsed.path) >= 2:
-                itemtype = parsed.path[1]
-                path = parsed.path[2:]
+                # we remove the selector
+                splitted = parsed.path.split("/")
+                # the first split if empty
+                if len(splitted[0]) == 0: splitted.pop(0)
+                #now we see if the element is a selector on not
+                if len(splitted[0]) == 1:
+                    path = parsed.path[2:]
+                else: 
+                    path = parsed.path
             else:
-                itemtype = "1"
                 path = ""
-            if itemtype == "0":
-                mime = "text/gemini"
-            elif itemtype == "1":
-                mime = "text/gopher"
-            elif itemtype == "h":
-                mime = "text/html"
-            elif itemtype in ("9", "g", "I", "s", ";"):
-                mime = "binary"
-            else:
-                mime = "text/gopher"
         else:
             path = parsed.path
         if parsed.query:
@@ -207,7 +213,11 @@ def get_cache_path(url, add_index=True):
     if local:
         cache_path = path
     elif scheme and host:
-        cache_path = os.path.expanduser(xdg("cache") + scheme + "/" + host + path)
+        # shemepath should not starts with / but finish with /
+        schemepath = ""
+        if include_protocol: schemepath = scheme + "/"
+        if subfolder: schemepath += subfolder + "/"
+        cache_path = os.path.expanduser(xdg(xdgfolder) + schemepath + host + path)
         # There’s an OS limitation of 260 characters per path.
         # We will thus cut the path enough to add the index afterward
         cache_path = cache_path[:249]
@@ -301,34 +311,51 @@ def set_error(url, err):
         if os.path.isdir(cache_dir):
             with open(cache, "w") as c:
                 c.write(str(datetime.datetime.now()) + "\n")
-                c.write("ERROR while caching %s\n\n" % url)
+                c.write(_("ERROR while caching %s\n\n") % url)
                 c.write("*****\n\n")
                 c.write(str(type(err)) + " = " + str(err))
                 # cache.write("\n" + str(err.with_traceback(None)))
                 c.write("\n*****\n\n")
-                c.write("If you believe this error was temporary, type " "reload" ".\n")
-                c.write("The resource will be tentatively fetched during next sync.\n")
+                c.write(_("If you believe this error was temporary, type " "reload" ".\n"))
+                c.write(_("The resource will be tentatively fetched during next sync.\n"))
                 c.close()
     return cache
 
 
+def get_cookiejar(url, create=False):
+    parsed = urllib.parse.urlparse(url)
+    basedir = os.path.join(xdg("data") + "cookies/" )
+    cookie_path = basedir + parsed.netloc + ".txt"
+    if not os.path.exists(cookie_path):
+        if create:
+            os.makedirs(basedir, exist_ok=True)
+            with open(cookie_path, 'w') as f:
+                f.write("# Netscape HTTP Cookie File\n\n")
+        else:
+            return None
+
+    jar = http.cookiejar.MozillaCookieJar(cookie_path)
+    jar.load()
+    return jar
+
 def _fetch_http(
     url,
     max_size=None,
     timeout=DEFAULT_TIMEOUT,
     accept_bad_ssl_certificates=False,
     force_large_download=False,
+    cookiejar=None,
     **kwargs,
 ):
     if not _DO_HTTP:
         return None
 
     def too_large_error(url, length, max_size):
-        err = "Size of %s is %s Mo\n" % (url, length)
-        err += "Offpunk only download automatically content under %s Mo\n" % (
+        err = _("Size of %s is %s Mo\n") % (url, length)
+        err += _("Offpunk only download automatically content under %s Mo\n") % (
             max_size / 1000000
         )
-        err += "To retrieve this content anyway, type 'reload'."
+        err += _("To retrieve this content anyway, type 'reload'.")
         return set_error(url, err)
 
     if accept_bad_ssl_certificates:
@@ -341,7 +368,8 @@ def _fetch_http(
     header = {}
     header["User-Agent"] = "Offpunk/Netcache - https://offpunk.net"
     with requests.get(
-        url, verify=verify, headers=header, stream=True, timeout=DEFAULT_TIMEOUT
+        url, verify=verify, headers=header, stream=True, timeout=DEFAULT_TIMEOUT,
+        cookies=cookiejar
     ) as response:
         if "content-type" in response.headers:
             mime = response.headers["content-type"]
@@ -367,30 +395,37 @@ def _fetch_http(
                 if current > downloaded:
                     downloaded = current
                     print(
-                        "  -> Receiving stream: %s%% of allowed data" % downloaded,
+                        _("  -> Receiving stream: %s%% of allowed data") % downloaded,
                         end="\r",
                     )
                 # print("size: %s (%s\% of maxlenght)"%(size,size/max_size))
                 if size > max_size / 2:
                     response.close()
                     return too_large_error(url, "streaming", max_size)
-            response.close()
         else:
             body = response.content
-            response.close()
+        if cookiejar is not None:
+            requests.cookies.extract_cookies_to_jar(cookiejar, response.request, response.raw)
+        response.close()
     if mime and "text/" in mime:
         body = body.decode("UTF-8", "replace")
     cache = write_body(url, body, mime)
     return cache
 
 
-def _fetch_gopher(url, timeout=DEFAULT_TIMEOUT, **kwargs):
+def _fetch_gopher(url, timeout=DEFAULT_TIMEOUT, interactive=True, **kwargs):
     parsed = urllib.parse.urlparse(url)
     host = parsed.hostname
     port = parsed.port or 70
     if len(parsed.path) >= 2:
         itemtype = parsed.path[1]
-        selector = parsed.path[2:]
+        # this is unreliable on "selectors" that contain "?" for example
+        # (which is perfectly valid)
+        # (urlparse took things after the ? as parsed.query)
+        # we should just get the full selector as is and use it
+        # selector = parsed.path[2:]
+        # so, we get everything after hostname:port, minus the 'itemtype' ([:1])
+        selector = url.split(parsed.netloc, 1)[1][2:]
     else:
         itemtype = "1"
         selector = ""
@@ -404,11 +439,17 @@ def _fetch_gopher(url, timeout=DEFAULT_TIMEOUT, **kwargs):
             break
         except OSError as e:
             err = e
-    if parsed.query:
-        request = selector + "\t" + parsed.query
-    elif itemtype == "7":
-        user_input = input("> ")
-        request = selector + "\t" + user_input
+    # gophermap lines can't have a query included.
+    # if there is something in parsed.query, it's because an error
+    # or a rogue "?" character in the selector
+    # if parsed.query:
+    #     request = selector + "\t" + parsed.query
+    if itemtype == "7":
+        if interactive:
+            user_input = input("> ")
+            request = selector + "\t" + user_input
+        else:
+            return None
     else:
         request = selector
     request += "\r\n"
@@ -520,11 +561,11 @@ def _validate_cert(address, host, cert, accept_bad_ssl=False, automatic_choice=N
         if accept_bad_ssl:
             if c.not_valid_before >= now:
                 raise CertificateError(
-                    "Certificate not valid until: {}!".format(c.not_valid_before)
+                    _("Certificate not valid until: {}!").format(c.not_valid_before)
                 )
             elif c.not_valid_after <= now:
                 raise CertificateError(
-                    "Certificate expired as of: {})!".format(c.not_valid_after)
+                    _("Certificate expired as of: {})!").format(c.not_valid_after)
                 )
 
         # Check certificate hostnames
@@ -553,7 +594,7 @@ def _validate_cert(address, host, cert, accept_bad_ssl=False, automatic_choice=N
         else:
             # If we didn't break out, none of the names were valid
             raise CertificateError(
-                "Hostname does not match certificate common name or any alternative names."
+                _("Hostname does not match certificate common name or any alternative names.")
             )
 
     sha = hashlib.sha256()
@@ -609,37 +650,38 @@ def _validate_cert(address, host, cert, accept_bad_ssl=False, automatic_choice=N
             print(previous_ttl)
 
         print("****************************************")
-        print("[SECURITY WARNING] Unrecognised certificate!")
+        print(_("[SECURITY WARNING] Unrecognised certificate!"))
         print(
-            "The certificate presented for {} ({}) has never been seen before.".format(
+            _("The certificate presented for {} ({}) has never been seen before.").format(
                 host, address
             )
         )
-        print("This MIGHT be a Man-in-the-Middle attack.")
+        print(_("This MIGHT be a Man-in-the-Middle attack."))
         print(
-            "A different certificate has previously been seen {} times.".format(
+            _("A different certificate has previously been seen {} times.").format(
                 max_count
             )
         )
         if _HAS_CRYPTOGRAPHY:
             if previous_ttl < datetime.timedelta():
-                print("That certificate has expired, which reduces suspicion somewhat.")
+                print(_("That certificate has expired, which reduces suspicion somewhat."))
             else:
-                print("That certificate is still valid for: {}".format(previous_ttl))
+                print(_("That certificate is still valid for: {}").format(previous_ttl))
         print("****************************************")
-        print("Attempt to verify the new certificate fingerprint out-of-band:")
+        print(_("Attempt to verify the new certificate fingerprint out-of-band:"))
         print(fingerprint)
         if automatic_choice:
             choice = automatic_choice
         else:
-            choice = input("Accept this new certificate? Y/N ").strip().lower()
+            #TRANSLATORS: keep "Y/N" because the answer has to be one of those
+            choice = input(_("Accept this new certificate? Y/N ")).strip().lower()
         if choice in ("y", "yes"):
             with open(os.path.join(sitedir, fingerprint), "w") as fp:
                 fp.write("1")
             with open(os.path.join(certcache, fingerprint + ".crt"), "wb") as fp:
                 fp.write(cert)
         else:
-            raise Exception("TOFU Failure!")
+            raise Exception(_("TOFU Failure!"))
 
     # 3. If no directory or no cert found in it, we cache it
     if not most_frequent_cert:
@@ -736,26 +778,27 @@ def create_certificate(name: str, days: int, hostname: str):
 def ask_certs(url: str):
     certs = get_certs(url)
     if len(certs) == 0:
-        print("There are no certificates available for this site.")
-        create_cert = input("Do you want to create one? (y/n) ")
+        print(_("There are no certificates available for this site."))
+        #TRANSLATORS: keep the "y/n"
+        create_cert = input(_("Do you want to create one? (y/n) "))
         if create_cert == "y":
-            name = input("Name for this certificate: ")
-            days = input("Validity in days: ")
+            name = input(_("Name for this certificate: "))
+            days = input(_("Validity in days: "))
             if name != "" and days.isdigit():
                 site = urllib.parse.urlparse(url)
                 create_certificate(name, int(days), site.hostname)
                 new_url = "gemini://" + name +"@"+ url.split("://")[1]
                 return(new_url)
             else:
-                print("The name or validity you typed are invalid")
+                print(_("The name or validity you typed are invalid"))
                 return(url)
         else:
             return(url)
     if len(certs) == 1:
-        print("The one available certificate for this site is:")
+        print(_("The one available certificate for this site is:"))
     elif len(certs) > 1:
         print(
-                "The", len(certs), "available certificates for this site are:"
+                _("The {} available certificates for this site are:").format(len(certs))
         )
     if len(certs) > 0:
         counter = 0
@@ -764,7 +807,7 @@ def ask_certs(url: str):
             stri += "[%s] %s \n" % (counter + 1, cert)
             counter += 1
         stri += "\n"
-        stri += "which certificate do you want to use? > "
+        stri += _("which certificate do you want to use? > ")
         ans = input(stri)
         if ans.isdigit() and 0 < int(ans) <= len(certs):
             identity = certs[int(ans) -1]
@@ -815,7 +858,7 @@ def _fetch_gemini(
     query = url_parts.query
     # In AV-98, this was the _send_request method
     # Send a selector to a given host and port.
-    # Returns the resolved address and binary file with the reply."""
+    # Returns the resolved address and binary file with the reply.
     host = host.encode("idna").decode()
     # Do DNS resolution
     # DNS lookup - will get IPv4 and IPv6 records if IPv6 is enabled
@@ -849,7 +892,7 @@ def _fetch_gemini(
         if certkey:
             context.load_cert_chain(certkey["cert"], certkey["key"])
         else:
-            print("This identity doesn't exist for this site (or is disabled).")
+            print(_("This identity doesn't exist for this site (or is disabled)."))
     # Impose minimum TLS version
     # In 3.7 and above, this is easy...
     if sys.version_info.minor >= 7:
@@ -914,13 +957,13 @@ def _fetch_gemini(
     header = f.readline(1027)
     header = urllib.parse.unquote(header.decode("UTF-8"))
     if not header or header[-1] != "\n":
-        raise RuntimeError("Received invalid header from server!")
+        raise RuntimeError(_("Received invalid header from server!"))
     header = header.strip()
     # Validate header
     status, meta = header.split(maxsplit=1)
     if len(meta) > 1024 or len(status) != 2 or not status.isnumeric():
         f.close()
-        raise RuntimeError("Received invalid header from server!")
+        raise RuntimeError(_("Received invalid header from server!"))
     # Update redirect loop/maze escaping state
     if not status.startswith("3"):
         previous_redirectors = set()
@@ -945,12 +988,12 @@ def _fetch_gemini(
     elif status.startswith("3"):
         newurl = urllib.parse.urljoin(url, meta)
         if newurl == url:
-            raise RuntimeError("URL redirects to itself!")
+            raise RuntimeError(_("URL redirects to itself!"))
         elif newurl in previous_redirectors:
-            raise RuntimeError("Caught in redirect loop!")
+            raise RuntimeError(_("Caught in redirect loop!"))
         elif len(previous_redirectors) == _MAX_REDIRECTS:
             raise RuntimeError(
-                "Refusing to follow more than %d consecutive redirects!"
+                _("Refusing to follow more than %d consecutive redirects!")
                 % _MAX_REDIRECTS
             )
         # TODO: redirections handling should be refactored
@@ -981,15 +1024,15 @@ def _fetch_gemini(
     # Client cert
     elif status.startswith("6"):
         if interactive:
-            print("You need to provide a client-certificate to access this page.")
+            print(_("You need to provide a client-certificate to access this page."))
             url_with_identity = ask_certs(url)
             if (url_with_identity != url):
                 return fetch(url_with_identity)
-        error = "You need to provide a client-certificate to access this page.\r\nType \"certs\" to create or re-use one"
+        error = _("You need to provide a client-certificate to access this page.\r\nType \"certs\" to create or re-use one")
         raise RuntimeError(error)
     # Invalid status
     elif not status.startswith("2"):
-        raise RuntimeError("Server returned undefined status code %s!" % status)
+        raise RuntimeError(_("Server returned undefined status code %s!") % status)
     # If we're here, this must be a success and there's a response body
     assert status.startswith("2")
     mime = meta
@@ -1013,8 +1056,8 @@ def _fetch_gemini(
             body = fbody.decode(encoding)
         except UnicodeError:
             raise RuntimeError(
-                "Could not decode response body using %s\
-                                encoding declared in header!"
+                _("Could not decode response body using %s\
+                                encoding declared in header!")
                 % encoding
             )
     else:
@@ -1022,19 +1065,47 @@ def _fetch_gemini(
     cache = write_body(url, body, mime)
     return cache, url
 
-
+#fetch returns two things:
+#cachepath: the path to the cached ressource
+#newurl: the real URL of that cached ressource
 def fetch(
     url,
     offline=False,
     download_image_first=True,
     images_mode="readable",
     validity=0,
+    cookiejar=None,
+    redirects={},
+    #blocked is empty by default to allow blocking rules having been removed
+    blocked={},
     **kwargs,
 ):
     url = normalize_url(url)
     newurl = url
     path = None
     print_error = "print_error" in kwargs.keys() and kwargs["print_error"]
+    #Step 0: we apply redirect and/or block
+    #Let’s add the blocked list into the redirects
+    #This will not overwrite existing rule for that domain
+    for b in blocked:
+        if b not in redirects.keys(): redirects[b] = "blocked"
+    redirection, key = get_url_redirected(url,redirects,returnkey=True)
+    if redirection and redirection.lower() == "blocked":
+        text = ""
+        text += _("Blocked URL: ")+url + "\n"
+        text += _("This website has been blocked with the following rule:\n")
+        text += key + "\n"
+        text += _("Use the following redirect command to unblock it:\n")
+        text += "redirect %s NONE" %key
+        if print_error:
+            print(text)
+        cache = set_error(newurl, text)
+        return cache, newurl
+    elif redirection:
+        parsed = urllib.parse.urlparse(url)
+        parsed = parsed._replace(netloc=redirection)
+        url = urllib.parse.urlunparse(parsed)
+        newurl = url
     # First, we look if we have a valid cache, even if offline
     # If we are offline, any cache is better than nothing
     if is_cache_valid(url, validity=validity) or (
@@ -1053,13 +1124,15 @@ def fetch(
             scheme = url.split("://")[0]
             if scheme not in standard_ports:
                 if print_error:
-                    print("%s is not a supported protocol" % scheme)
+                    print(_("%s is not a supported protocol") % scheme)
                 path = None
             elif scheme in ("http", "https"):
                 if _DO_HTTP:
-                    path = _fetch_http(newurl, **kwargs)
+                    if download_image_first and cookiejar is None:
+                        cookiejar = get_cookiejar(newurl)
+                    path = _fetch_http(newurl, cookiejar=cookiejar, **kwargs)
                 else:
-                    print("HTTP requires python-requests")
+                    print(_("HTTP requires python-requests"))
             elif scheme == "gopher":
                 path = _fetch_gopher(newurl, **kwargs)
             elif scheme == "finger":
@@ -1078,45 +1151,45 @@ def fetch(
             # we fail silently when sync_only
             if isinstance(err, socket.gaierror):
                 if print_error:
-                    print("ERROR: DNS error!")
+                    print(_("ERROR: DNS error!"))
             elif isinstance(err, ConnectionRefusedError):
                 if print_error:
-                    print("ERROR1: Connection refused!")
+                    print(_("ERROR1: Connection refused!"))
             elif isinstance(err, ConnectionResetError):
                 if print_error:
-                    print("ERROR2: Connection reset!")
+                    print(_("ERROR2: Connection reset!"))
             elif isinstance(err, (TimeoutError, socket.timeout)):
                 if print_error:
-                    print("""ERROR3: Connection timed out!
-    Slow internet connection?  Use 'set timeout' to be more patient.""")
+                    print(_("""ERROR3: Connection timed out!
+    Slow internet connection?  Use 'set timeout' to be more patient."""))
             elif isinstance(err, FileExistsError):
                 if print_error:
-                    print("""ERROR5: Trying to create a directory which already exists
-                        in the cache : """)
+                    print(_("""ERROR5: Trying to create a directory which already exists
+                        in the cache : """))
                 print(err)
             elif _DO_HTTP and isinstance(err, requests.exceptions.SSLError):
                 if print_error:
-                    print("""ERROR6: Bad SSL certificate:\n""")
+                    print(_("""ERROR6: Bad SSL certificate:\n"""))
                     print(err)
                     print(
-                        """\n If you know what you are doing, you can try to accept bad certificates with the following command:\n"""
+                        _("""\n If you know what you are doing, you can try to accept bad certificates with the following command:\n""")
                     )
                     print("""set accept_bad_ssl_certificates True""")
             elif _DO_HTTP and isinstance(err, requests.exceptions.ConnectionError):
                 if print_error:
-                    print("""ERROR7: Cannot connect to URL:\n""")
+                    print(_("""ERROR7: Cannot connect to URL:\n"""))
                     print(str(err))
             else:
                 if print_error:
                     import traceback
 
-                    print("ERROR4: " + str(type(err)) + " : " + str(err))
+                    print(_("ERROR4: ") + str(type(err)) + " : " + str(err))
                     # print("\n" + str(err.with_traceback(None)))
                     print(traceback.format_exc())
             return cache, newurl
         # We download images contained in the document (from full mode)
         if not offline and download_image_first and images_mode:
-            renderer = ansicat.renderer_from_file(path, newurl)
+            renderer = ansicat.renderer_from_file(path, newurl,redirectlist=redirects,**kwargs)
             if renderer:
                 for image in renderer.get_images(mode=images_mode):
                     # Image should exist, should be an url (not a data image)
@@ -1127,7 +1200,7 @@ def fetch(
                         and not is_cache_valid(image)
                     ):
                         width = offutils.term_width() - 1
-                        toprint = "Downloading %s" % image
+                        toprint = _("Downloading %s") % image
                         toprint = toprint[:width]
                         toprint += " " * (width - len(toprint))
                         print(toprint, end="\r")
@@ -1139,56 +1212,60 @@ def fetch(
                             download_image_first=False,
                             images_mode=None,
                             validity=0,
+                            cookiejar=cookiejar,
+                            redirects=redirects,
                             **kwargs,
                         )
+    if download_image_first and cookiejar is not None:
+        cookiejar.save()
     return path, newurl
 
 
 def main():
-    descri = "Netcache is a command-line tool to retrieve, cache and access networked content.\n\
+    descri = _("Netcache is a command-line tool to retrieve, cache and access networked content.\n\
             By default, netcache will returns a cached version of a given URL, downloading it \
             only if a cache version doesn't exist. A validity duration, in seconds, can also \
-            be given so netcache downloads the content only if the existing cache is older than the validity."
+            be given so netcache downloads the content only if the existing cache is older than the validity.")
     # Parse arguments
     parser = argparse.ArgumentParser(prog="netcache", description=descri)
     parser.add_argument(
         "--path",
         action="store_true",
-        help="return path to the cache instead of the content of the cache",
+        help=_("return path to the cache instead of the content of the cache"),
     )
     parser.add_argument(
         "--ids",
         action="store_true",
-        help="return a list of id's for the gemini-site instead of the content of the cache",
+        help=_("return a list of id's for the gemini-site instead of the content of the cache"),
     )
     parser.add_argument(
         "--offline",
         action="store_true",
-        help="Do not attempt to download, return cached version or error",
+        help=_("Do not attempt to download, return cached version or error"),
     )
     parser.add_argument(
         "--max-size",
         type=int,
-        help="Cancel download of items above that size (value in Mb).",
+        help=_("Cancel download of items above that size (value in Mb)."),
     )
     parser.add_argument(
         "--timeout",
         type=int,
-        help="Time to wait before cancelling connection (in second).",
+        help=_("Time to wait before cancelling connection (in second)."),
     )
     parser.add_argument(
         "--cache-validity",
         type=int,
         default=0,
-        help="maximum age, in second, of the cached version before \
-                                redownloading a new version",
+        help=_("maximum age, in second, of the cached version before \
+                                redownloading a new version"),
     )
     # No argument: write help
     parser.add_argument(
         "url",
         metavar="URL",
         nargs="*",
-        help="download URL and returns the content or the path to a cached version",
+        help=_("download URL and returns the content or the path to a cached version"),
     )
     # --validity : returns the date of the cached version, Null if no version
     # --force-download : download and replace cache, even if valid
@@ -1206,6 +1283,8 @@ def main():
                 max_size=args.max_size,
                 timeout=args.timeout,
                 validity=args.cache_validity,
+                redirects=offblocklist.redirects,
+                blocked=offblocklist.blocked,
             )
         if args.path:
             print(path)
