@@ -1,5 +1,5 @@
 #!/usr/bin/env python3
-# opnk stand for "Open like a PuNK".
+# openk stand for "Open like a PuNK".
 # It will open any file or URL and display it nicely in less.
 # If not possible, it will fallback to xdg-open
 # URL are retrieved through netcache
@@ -10,20 +10,35 @@ import shutil
 import sys
 import tempfile
 import time
+import gettext
 
 import ansicat
 import netcache
 import offutils
-from offutils import GREPCMD, is_local, mode_url, run, term_width, unmode_url, init_config
+import offblocklist
+from offutils import (
+        GREPCMD, 
+        is_local, 
+        mode_url, 
+        run, 
+        term_width, 
+        unmode_url, 
+        init_config,
+        send_email,
+        _HAS_XDGOPEN,
+        _LOCALE_DIR,
+        )
 
-_HAS_XDGOPEN = shutil.which("xdg-open")
+gettext.bindtextdomain('offpunk', _LOCALE_DIR)
+gettext.textdomain('offpunk')
+_ = gettext.gettext
 
 less_version = 0
 if not shutil.which("less"):
-    print('Please install the pager "less" to run Offpunk.')
-    print("If you wish to use another pager, send me an email !")
+    print(_('Please install the pager "less" to run Offpunk.'))
+    print(_("If you wish to use another pager, send me an email !"))
     print(
-        '(I’m really curious to hear about people not having "less" on their system.)'
+        _('(I’m really curious to hear about people not having "less" on their system.)')
     )
     sys.exit()
 output = run("less --version")
@@ -97,6 +112,7 @@ class opencache:
         self.mime_handlers = {}
         self.last_mode = {}
         self.last_width = term_width(absolute=True)
+        self.redirects = offblocklist.redirects
 
     def _get_handler_cmd(self, mimetype,file_extension=None):
         # Now look for a handler for this mimetype
@@ -119,8 +135,9 @@ class opencache:
             if _HAS_XDGOPEN:
                 cmd_str = "xdg-open %s"
             else:
-                cmd_str = 'echo "Can’t find how to open "%s'
-                print("Please install xdg-open (usually from xdg-util package)")
+                #TRANSLATORS: keep echo and %s, translate the text between ""
+                cmd_str = _('echo "Can’t find how to open "%s')
+                print(_("Please install xdg-open (usually from xdg-util package)"))
         return cmd_str
 
     # Return the handler for a specific mimetype.
@@ -164,7 +181,6 @@ class opencache:
             if usecache and self.last_width != width:
                 self.cleanup()
                 usecache = False
-                self.last_width = width
             if usecache:
                 if inpath in self.renderer_time.keys():
                     last_downloaded = netcache.cache_last_modified(inpath)
@@ -177,7 +193,7 @@ class opencache:
                     usecache = False
             if not usecache:
                 renderer = ansicat.renderer_from_file(path, url=inpath, theme=theme,\
-                                                      **kwargs)
+                                                      redirectlist=self.redirects,**kwargs)
                 if renderer:
                     self.rendererdic[inpath] = renderer
                     self.renderer_time[inpath] = int(time.time())
@@ -191,8 +207,8 @@ class opencache:
         else:
             return None
 
-    def opnk(self, inpath, mode="readable", terminal=True, grep=None, theme=None, link=None,\
-                        direct_open_unsupported=False, **kwargs):
+    def openk(self, inpath, mode="readable", terminal=True, grep=None, theme=None, \
+                link=None, direct_open_unsupported=False, **kwargs):
         # Return True if inpath opened in Terminal
         # False otherwise
         # also returns the url in case it has been modified
@@ -205,18 +221,18 @@ class opencache:
         if not offutils.is_local(inpath):
             if mode:
                 kwargs["images_mode"] = mode
-            cachepath, inpath = netcache.fetch(inpath, **kwargs)
+            cachepath, inpath = netcache.fetch(inpath, redirects=self.redirects,**kwargs)
             if not cachepath:
                 return False, inpath
         # folowing line is for :// which are locals (file,list)
         elif "://" in inpath:
-            cachepath, inpath = netcache.fetch(inpath, **kwargs)
+            cachepath, inpath = netcache.fetch(inpath, redirects=self.redirects,**kwargs)
         elif inpath.startswith("mailto:"):
             cachepath = inpath
         elif os.path.exists(inpath):
             cachepath = inpath
         else:
-            print("%s does not exist" % inpath)
+            print(_("%s does not exist") % inpath)
             return False, inpath
         renderer = self.get_renderer(inpath, mode=mode, theme=theme, **kwargs)
         if link and link.isdigit():
@@ -285,23 +301,18 @@ class opencache:
                 extension = last_part.split(".")[-1]
             if mimetype == "mailto":
                 mail = inpath[7:]
-                resp = input("Send an email to %s Y/N? " % mail)
-                if resp.strip().lower() in ("y", "yes"):
-                    if _HAS_XDGOPEN:
-                        run("xdg-open mailto:%s", parameter=mail, direct_output=True)
-                    else:
-                        print("Cannot find a mail client to send mail to %s" % inpath)
-                        print("Please install xdg-open (usually from xdg-util package)")
+                send_email(mail,toconfirm=True)
                 return False, inpath
             else:
                 cmd_str = self._get_handler_cmd(mimetype,file_extension=extension)
-            change_cmd = "\"handler %s MY_PREFERED_APP %%s\""%mimetype
+            #TRANSLATORS translate only "MY_PREFERED_APP"
+            change_cmd = _("\"handler %s MY_PREFERED_APP %%s\"")%mimetype
             try:
                 #we don’t write the info if directly opening to avoid 
-                #being verbose in opnk
+                #being verbose in openk
                 if not direct_open_unsupported:
-                    print("External open of type %s with \"%s\""%(mimetype,cmd_str))
-                    print("You can change the default handler with %s"%change_cmd)
+                    print(_("External open of type %s with \"%s\"")%(mimetype,cmd_str))
+                    print(_("You can change the default handler with %s")%change_cmd)
                 run(
                     cmd_str,
                     parameter=netcache.get_cache_path(inpath),
@@ -309,11 +320,11 @@ class opencache:
                 )
                 return True, inpath
             except FileNotFoundError:
-                print("Handler program %s not found!" % shlex.split(cmd_str)[0])
-                print("You can use the ! command to specify another handler program\
-                        or pipeline.")
+                print(_("Handler program %s not found!") % shlex.split(cmd_str)[0])
+                print(_("You can use the ! command to specify another handler program\
+                        or pipeline."))
 
-                print("You can change the default handler with %s"%change_cmd)
+                print(_("You can change the default handler with %s")%change_cmd)
             return False, inpath
 
     # We remove the renderers from the cache and we also delete temp files
@@ -322,7 +333,8 @@ class opencache:
             os.remove(self.temp_files.popitem()[1])
         while len(self.less_histfile) > 0:
             os.remove(self.less_histfile.popitem()[1])
-        self.last_width = None
+        #After cleanup, we set the current size of the terminal
+        self.last_width = term_width(absolute=True)
         self.rendererdic = {}
         self.renderer_time = {}
         self.last_mode = {}
@@ -335,8 +347,11 @@ class opencache:
         # First, we take the full URL
         if mode:
             url = mode_url(url,mode)
-        delfile(self.temp_files.pop(url))
-        delfile(self.less_histfile.pop(url))
+        # Unrendered element, such as picture, are not in the dictionnaries
+        if url in self.temp_files:
+            delfile(self.temp_files.pop(url))
+        if url in self.less_histfile:
+            delfile(self.less_histfile.pop(url))
         url, newmode = unmode_url(url)
         if url in self.rendererdic:
             self.rendererdic.pop(url)
@@ -345,30 +360,38 @@ class opencache:
 
 
 def main():
-    descri = "opnk is an universal open command tool that will try to display any file \
+    descri = _("openk is an universal open command tool that will try to display any file \
              in the pager less after rendering its content with ansicat. If that fails, \
-             opnk will fallback to opening the file with xdg-open. If given an URL as input \
-             instead of a path, opnk will rely on netcache to get the networked content."
-    parser = argparse.ArgumentParser(prog="opnk", description=descri)
+             openk will fallback to opening the file with xdg-open. If given an URL as input \
+             instead of a path, openk will rely on netcache to get the networked content.")
+    parser = argparse.ArgumentParser(prog="openk", description=descri)
     parser.add_argument(
         "--mode",
         metavar="MODE",
-        help="Which mode should be used to render: normal (default), full or source.\
-                                With HTML, the normal mode try to extract the article.",
+        help=_("Which mode should be used to render: normal (default), full or source.\
+                                With HTML, the normal mode try to extract the article."),
+    )
+    parser.add_argument(
+        "--linkmode",
+        choices=[
+            "none",
+            "end",
+        ],
+        help=_("Which mode should be used to render links: none (default) or end"),
     )
     parser.add_argument(
         "content",
         metavar="INPUT",
         nargs="*",
         default=sys.stdin,
-        help="Path to the file or URL to open",
+        help=_("Path to the file or URL to open"),
     )
     parser.add_argument(
         "--cache-validity",
         type=int,
         default=0,
-        help="maximum age, in second, of the cached version before \
-                                redownloading a new version",
+        help=_("maximum age, in second, of the cached version before \
+                                redownloading a new version"),
     )
     args = parser.parse_args()
     cache = opencache()
@@ -380,15 +403,18 @@ def main():
             cache.set_handler(splitted[1],splitted[2])
     # if the second argument is an integer, we associate it with the previous url
     # to use as a link_id
-    if len(args.content) == 2 and args.content[1].isdigit():
+    if (type(args.content) == list) and len(args.content) == 2 and args.content[1].isdigit():
         url = args.content[0]
         link_id = args.content[1]
-        cache.opnk(url, mode=args.mode, validity=args.cache_validity, link=link_id,\
-                    direct_open_unsupported=True)
+        cache.openk(url, mode=args.mode, validity=args.cache_validity, link=link_id,\
+                    direct_open_unsupported=True, linkmode=args.linkmode)
     else:
         for f in args.content:
-            cache.opnk(f, mode=args.mode, validity=args.cache_validity,\
-                        direct_open_unsupported=True)
+            cache.openk(f, mode=args.mode, validity=args.cache_validity,\
+                        direct_open_unsupported=True, linkmode=args.linkmode)
 
 if __name__ == "__main__":
+    if "opnk" in sys.argv[0]:
+        print("WARNING: opnk.py has been deprecated in favour of openk.py")
+        print("*******  Replace all your opnk calls by \"openk\"")
     main()
